﻿// This file contains your Data Connector logic
[Version = "1.0.0"]
section ZetarisDataFabric;

EnableTraceOutput = true;

[DataSource.Kind="ZetarisDataFabric", Publish="ZetarisDataFabric.Publish"]
shared ZetarisDataFabric.Database = (server as text, port as number, optional certification_path as text ) as table =>
    let
		Credential = Extension.CurrentCredential(),
        ConnectionString = [
            Driver = "Zetaris Networked Data Platform",
            SERVER = server,
            PORT = port,
			UID = Credential[Username],
			PWD = Credential[Password],
            TSTPATH = if (certification_path <> null) then certification_path else "",
            SSL = if (certification_path <> null) then 1 else 0
        ],

        SQLGetTypeInfo = (types) => 
            if (EnableTraceOutput <> true) then types else
            let
                // Outputting the entire table might be too large, and result in the value being truncated.
                // We can output a row at a time instead with Table.TransformRows()
                rows = Table.TransformRows(types, each Diagnostics.LogValue("SQLGetTypeInfo " & _[TYPE_NAME], _)),
                toTable = Table.FromRecords(rows)
            in
                Value.ReplaceType(toTable, Value.Type(types)),   

        SQLColumns = (catalogName, schemaName, tableName, columnName, source) =>
            if (EnableTraceOutput <> true) then source else
            // the if statement conditions will force the values to evaluated/written to diagnostics
            if (Diagnostics.LogValue("SQLColumns.TableName", tableName) <> "***" and Diagnostics.LogValue("SQLColumns.ColumnName", columnName) <> "***") then
                let
                    // Outputting the entire table might be too large, and result in the value being truncated.
                    // We can output a row at a time instead with Table.TransformRows()
                    rows = Table.TransformRows(source, each Diagnostics.LogValue("SQLColumns", _)),
                    toTable = Table.FromRecords(rows)
                in
                    Value.ReplaceType(toTable, Value.Type(source))
            else
                source,
		
		OdbcDataSource = Odbc.DataSource(ConnectionString, [
            // Enables client side connection pooling for the ODBC driver.
            // Most drivers will want to set this value to true.
            ClientConnectionPooling = true,
            // When HierarchialNavigation is set to true, the navigation tree
            // will be organized by Database -> Schema -> Table. When set to false,
            // all tables will be displayed in a flat list using fully qualified names. 
            HierarchicalNavigation = true,
            // Use the SqlCapabilities record to specify driver capabilities that are not
            // discoverable through ODBC 3.8, and to override capabilities reported by
            // the driver. 
            SqlCapabilities = [
                GroupByCapabilities =  ODBC[SQL_GB][SQL_GB_GROUP_BY_CONTAINS_SELECT],
                FractionalSecondsScale = 3,

                Sql92Conformance = ODBC[SQL_SC][SQL_SC_SQL92_FULL],
                //GroupByCapabilities = SQL_GB[SQL_GB_COLLATE],
                
                SupportsNumericLiterals = true,
                SupportsStringLiterals = true,               
                SupportsOdbcDateLiterals = true,
                SupportsOdbcTimeLiterals = true,
                SupportsOdbcTimestampLiterals = true,
                StringLiteralEscapeCharacters = {"\"},
                SupportsTop = false,
                LimitClauseKind = LimitClauseKind.Limit,
                PrepareStatements = false
            ],
            
            SQLGetFunctions = [
                SQL_API_SQLBINDPARAMETER = false,
                SQL_CONVERT_FUNCTIONS = 0x2 /* SQL_FN_CVT_CAST */
            ],
            SQLGetInfo = [
                SQL_SQL_CONFORMANCE = SQL_SC[SQL_SC_SQL92_FULL],  // null, 1, 2, 4, 8,
                SQL_SQL92_PREDICATES = ODBC[SQL_SP][All],
                SQL_AGGREGATE_FUNCTIONS = ODBC[SQL_AF][All],
                SQL_GROUP_BY = ODBC[SQL_GB][SQL_GB_GROUP_BY_CONTAINS_SELECT]
            ],
    
            AstVisitor = [
                LimitClause = (skip, take) =>
                if (skip > 0) then error "Skip/Offset not supported"
                else if (take <> null) then
                [
                    Text = Text.Format("LIMIT #{0}", {take}),
                    Location = "AfterQuerySpecification"
                ]
                else ""
            ],
            
            SoftNumbers = true,
            //HideNativeQuery = true,
            SQLColumns = SQLColumns,
            SQLGetTypeInfo = SQLGetTypeInfo
            
		])
    in
        OdbcDataSource;


// Data Source Kind description
ZetarisDataFabric = [
    Authentication = [
        UsernamePassword = []
    ],
    Label = "Direct Query for Zetaris Data Fabric"
];

// Data Source UI publishing description
ZetarisDataFabric.Publish = [
    Beta = true,
    SupportsDirectQuery = true,     // enables direct query
    Category = "Database",
    ButtonText = { "Zetaris Data Fabric", "Direct Query via ODBC for Zetaris Data Fabric" },
    SourceImage = ZetarisDataFabric.Icons,
    SourceTypeImage = ZetarisDataFabric.Icons
];

ZetarisDataFabric.Icons = [
    Icon16 = { Extension.Contents("ZetarisDataFabric16.png"), Extension.Contents("ZetarisDataFabric20.png"), Extension.Contents("ZetarisDataFabric24.png"), Extension.Contents("ZetarisDataFabric32.png") },
    Icon32 = { Extension.Contents("ZetarisDataFabric32.png"), Extension.Contents("ZetarisDataFabric40.png"), Extension.Contents("ZetarisDataFabric48.png"), Extension.Contents("ZetarisDataFabric64.png") }
];

Extension.LoadFunction = (name as text) =>
    let
        binary = Extension.Contents(name),
        asText = Text.FromBinary(binary)
    in
        Expression.Evaluate(asText, #shared);

Diagnostics = Extension.LoadFunction("Diagnostics.pqm");
Diagnostics.LogValue = if (EnableTraceOutput) then Diagnostics[LogValue] else (prefix, value) => value;

// OdbcConstants contains numeric constants from the ODBC header files, and a 
// helper function to create bitfield values.
ODBC = Extension.LoadFunction("OdbcConstants.pqm");

// Expose the constants and bitfield helpers
SQL_SC = ODBC[SQL_SC];
SQL_GB = ODBC[SQL_GB];


